#!/usr/bin/perl -w

use strict;
use warnings;


sub batch_filter_phrase_table {
    my($data,$parameters,$experiment_dir,$src_tokens,$first_sentence,$last_sentence)=@_;
    
    my $phrase_table=$$data{'phrase_table'};
    my $oister_home=$$data{'oister_home'};

    my $phrase_table_full_path=$phrase_table;
    $phrase_table=~s/\.gz$//o;
    $phrase_table=~s/^.*?\/?([^\/]+)$/$1/o;
    $oister_home=$$data{'oister_home'};

    my $filtered_phrase_table="$experiment_dir/filtered_phrase_tables/$phrase_table.$first_sentence-$last_sentence";

    if(-e "$filtered_phrase_table") {
	print STDERR "Filtered phrase table \"$filtered_phrase_table\" exists.\n";
	return $filtered_phrase_table;
    } elsif(-e "$filtered_phrase_table.trg_sorted") {
	print STDERR "Filtered phrase table \"$filtered_phrase_table.trg_sorted\" exists.\n";
	return "$filtered_phrase_table.trg_sorted";
    } elsif(-e "$filtered_phrase_table.gz") {
	print STDERR "Filtered phrase table \"$filtered_phrase_table.gz\" exists.\n";
	return "$filtered_phrase_table.gz";
    } elsif(-e "$filtered_phrase_table.trg_sorted.gz") {
	print STDERR "Filtered phrase table \"$filtered_phrase_table.trg_sorted.gz\" exists.\n";
	return "$filtered_phrase_table.trg_sorted.gz";
    }

    my $max_phrase_length=$$parameters{'max_phrase_length'};
    my %match_phrase;

    for(my $sent_id=$first_sentence; $sent_id<=$last_sentence && $sent_id<@$src_tokens; $sent_id++) {

	for(my $i=0; $i<@{ $src_tokens->[$sent_id] }; $i++) {
	    for(my $j=$i; $j<@{ $src_tokens->[$sent_id] } && $j<$i+$max_phrase_length; $j++) {
		my $phrase=join(" ",@{ $src_tokens->[$sent_id] }[$i..$j]);
		$match_phrase{$phrase}=1;
	    }
	}
    }


    my $no_lines_processed=0;
    my $prev_no_lines=0;
   
    my $prev_f='Z_#';
    my $match;
    print STDERR "Filtering phrase_table $phrase_table -> $filtered_phrase_table ... ";
    if($phrase_table_full_path=~/\.gz$/o) {
        open P, "<:gzip", $phrase_table_full_path, or die("can't open $phrase_table_full_path: $!\n");
    } else {
	open(P,"<$phrase_table_full_path") || die("can't open file $phrase_table_full_path: $!\n");
    }
    open(F,">$filtered_phrase_table.progress") || die("can't open file $filtered_phrase_table: $!\n");
    while(defined(my $line=<P>)) {
	my($f)=$line=~/^(.+?) \|\|\| /o;

	$no_lines_processed++;
	my $no_digits = rindex($prev_no_lines,"");
	if($no_lines_processed % 1000 == 0) {
	    for(my $j=1; $j<=$no_digits; $j++) {
		print STDERR "\x08";
	    };
	    print STDERR "$no_lines_processed";
            $prev_no_lines = $no_lines_processed;
	};
    

	if($f eq $prev_f) {
	    if($match) {
		print F $line;
		next;
	    } 
	} elsif(exists($match_phrase{$f})) {
		print F $line;
		$match=1;
	} else {
	    $match=0;
	}

	$prev_f=$f;	
    }
    close(P);
    close(F);

    print STDERR " done.\n";

    my $tmp_sort="$experiment_dir/tmp";

    print STDERR "cat $filtered_phrase_table.progress | $oister_home/scripts/phrase-table-swap-trg-src-sides.pl | sort -T $tmp_sort | gzip > $filtered_phrase_table.trg_sorted.gz ...";
    system("cat $filtered_phrase_table.progress | $oister_home/scripts/phrase-table-swap-trg-src-sides.pl | sort -T $tmp_sort | gzip > $filtered_phrase_table.trg_sorted.gz");
    unlink("$filtered_phrase_table.progress");
    return "$filtered_phrase_table.trg_sorted.gz";
}


return 1;
