
sub load_config_file {
    my($config_file,$line);

    ($config_file,$mert_mode,$data,$parameters,$features,$feature_order,$default_values,$rescore_lattice)=@_;

    my $feature_counter=0;

    open(F,"<$config_file")||die("can't open config_file $config_file: $!\n");
    while(defined($line=<F>)) {
	$line=~s/[\s\t]*(.*?)[\s\t]*\n/$1/;

	if($line=~/^\#/) {
	    #skip
	} elsif($line=~/^([^\:]+)\:([^\=]+)\=([^ ]+)(.*)$/) {
	    my $type=$1;
	    my $name=$2;
	    my $value=$3;	
	    my $mert_value=$4;
	    my $default_value;
	    $name=~s/\..+$//;

	    if($value=~/^([0-9\.\-e]+)\(([0-9\.\-e]+)\)/) {
		$value=$1;
		$default_value=$2;
	    }
	    if($type eq 'feature' && $mert_mode && defined($mert_value) && $mert_value=~/opt=([^\[]+)\[/) {
		$mert_value=~s/^.*[\s\t]opt=([^\[]+)\[.*$/$1/;
		$value=$mert_value;
	    } elsif($type eq 'feature' && $mert_mode) {
		print STDERR "mert opt value missing in config-file=$config_file\n";
		exit(-1);
	    }

	    if($type eq 'parameter') {
		$$parameters{$name}=$value;
	    } elsif($type eq 'feature') {
		if($name=~/phrase_table\[([0-9]+)\]/) {
		    my $phrase_table_index=$1;
		    $$features{'phrase_table'}[$phrase_table_index]=$value;
		    $$default_values{'phrase_table'}[$phrase_table_index]=$default_value if(defined($default_value));
		    print STDERR "default_values{'phrase_table'}[$phrase_table_index]=$$default_values{'phrase_table'}[$phrase_table_index]\n";
		    $$feature_order{$name}=$feature_counter++;
		    print STDERR "feature_order{$name}=$$feature_order{$name}\n";
		} elsif($name=~/lex_dm\[([0-9]+)\]/) {
		    my $dm_index=$1;
		    $$features{'lex_dm'}[$dm_index]=$value;
		    $$default_values{'lex_dm'}[$dm_index]=$default_value if(defined($default_value));
		    print STDERR "default_values{'lex_dm'}[$dm_index]=$$default_values{'lex_dm'}[$dm_index]\n" if(defined($default_value));
		    $$feature_order{$name}=$feature_counter++;
		    print STDERR "feature_order{$name}=$$feature_order{$name}\n";
		} else {
		    $$features{$name}=$value;
		    $$feature_order{$name}=$feature_counter++;
		    print STDERR "feature_order{$name}=$$feature_order{$name}\n";
		    print STDERR "features{$name}=$value\n";
		}
	    } elsif($type eq 'data') {
		$$data{$name}=$value;
	    } else {
		print STDERR "Incorrect configuration: $line\n";
	    }

	}

    }

    #default config parameter values:
    $$parameters{'poslm_order'}=0 if(!defined($$parameters{'poslm_order'}));
    $$parameters{'min_prune_stack_size'}=0 if(!defined($$parameters{'min_prune_stack_size'}));
    $$parameters{'max_phrase_length'}=7 if(!defined($$parameters{'max_phrase_length'}));
    $$parameters{'phrase_table_threshold'}=0 if(!defined($$parameters{'phrase_table_threshold'}));


    foreach my $name (keys %$parameters) {
	print STDERR "parameters{$name}=$$parameters{$name}\n";
    }

    # if in rescore_lattice mode, switch off all pruning
    if($rescore_lattice) {
	print STDERR "In lattice rescoring mode:\n";
	print STDERR "parameters{'lm_order'}=parameters{'rescore_lm_order'}=$$parameters{'rescore_lm_order'}\n";
	$$parameters{'lm_order'}=$$parameters{'rescore_lm_order'};
	print STDERR "parameters{'beam_width'}=0.0000000001\n";
	$$parameters{'beam_width'}=0.0000000001;
        print STDERR "parameters{'children_beam_width'}=0.0000000001\n";	$$parameters{'children_beam_width'}=0.0000000001;
	print STDERR "parameters{'stack_size'}=1000\n";
	$$parameters{'stack_size'}=1000;
    }
    
    close(F);
}


sub buffer_config_file {
    my($config_file)=@_;

    my @buffer;
    open(F,"<$config_file")||die("can't open config_file $config_file: $!\n");
    while(defined($line=<F>)) {
	push(@buffer,$line);
    }
    close(F);

    return @buffer;
}


sub return_initial_lambda_string {
    my($config_file)=@_;

    my @lambdas;
    open(F,"<$config_file")||die("can't open config_file $config_file: $!\n");
    while(defined($line=<F>)) {
	if($line=~/^feature:.+[\s\t]init=([^\[]+)\[([0-9\.e\-]+)\,([0-9\.e\-]+)\]/) {
	    my $value=$1;
	    my $lower=$2;
	    my $higher=$3;
	    push(@lambdas,"$value\,$lower\-$higher");
	}
    }
    close(F);

    return join(";",@lambdas);       
}

sub return_last_lambda_string {
    my($config_file)=@_;

    my @lambdas;
    open(F,"<$config_file")||die("can't open config_file $config_file: $!\n");
    while(defined($line=<F>)) {
	if($line=~/^feature:.+[\s\t]opt=([^\[]+)\[([0-9\.e\-]+)\,([0-9\.e\-]+)\]/) {
	    my $value=$1;
	    my $lower=$2;
	    my $higher=$3;
	    push(@lambdas,"$value\,$lower\-$higher");
	}
    }
    close(F);

    return join(";",@lambdas);       
}



return 1;
