

sub load_lexicalized_distortion_model {

    my($lex_dm_file,$lex_dm,$f2e_lex,$parameters,$features,$rescore_mode)=@_;


    my @lex_dm_feature_weights=@{ $$features{'lex_dm'} };
    my $unknown_lex_dm_prob=$$parameters{'unknown_lex_dm_prob'};

#    print STDERR "lex_dm_feature_weights=", @lex_dm_feature_weights, "\n";
#    die;

    my $direction;
    my @orientations;

    if($lex_dm_file=~/\.gz$/o) {
        open DM, "<:gzip", $lex_dm_file, or die("can't open $lex_dm_file: $!\n");
    } else {
        open(DM,"<$lex_dm_file") || die("can't open $lex_dm_file: $!\n");
    }

    # read in first line separately for UNKs
    my $line=<DM>;
    chomp($line);
    my($f_phrase,$e_phrase,$dm_score_string)=split(/ \|\|\| /,$line);
    my @dm_scores=split(/ /,$dm_score_string);

    if(@dm_scores==6) {
	$direction='bi';
	@orientations=qw( pm ps pd nm ns nd );
    } else {
	$direction='uni';
	@orientations=qw( pm ps pd );
    }


    if($rescore_mode) {
	for(my $i=0; $i<@dm_scores; $i++) {
	    if($dm_scores[$i]==0) {
		$$lex_dm{$f_phrase}{$e_phrase}[$i]=$unknown_lex_dm_prob;
	    } else {
		$$lex_dm{$f_phrase}{$e_phrase}[$i]=log($dm_scores[$i]);
	    }
	}
    } else {
	for(my $i=0; $i<@dm_scores; $i++) {
	    if($dm_scores[$i]==0) {
		$$lex_dm{$f_phrase}{$e_phrase}[$i]=$unknown_lex_dm_prob*$lex_dm_feature_weights[$i];
	    } else {
		$$lex_dm{$f_phrase}{$e_phrase}[$i]=log($dm_scores[$i])*$lex_dm_feature_weights[$i];
	    }
	}
    }

    # now continue with the rest of the DM file
    my $prev_no_lines=0;
    my $no_lines_processed=0;
    print STDERR "Reading in lexicalized distortion model \'$lex_dm_file\' ...  ";
    my $time_begin=time;
    while(defined(my $line=<DM>)) {

	$no_lines_processed++;
	my $no_digits = rindex($prev_no_lines,"");
	if($no_lines_processed % 100000 == 0) {
	    for(my $j=1; $j<=$no_digits; $j++) {
		print STDERR "\x08";
	    };
	    print STDERR $no_lines_processed;
	    $prev_no_lines = $no_lines_processed;
	}    

	my($f_phrase,$e_phrase,$dm_score_string)=split(/ \|\|\| /,$line);
	next if($f_phrase!~/^<\/?s>/o && (!exists($$f2e_lex{$f_phrase}) || !exists($$f2e_lex{$f_phrase}{$e_phrase})));

	my @dm_scores=split(/ /,$dm_score_string);

	if($rescore_mode) {
	    for(my $i=0; $i<@dm_scores; $i++) {
		if($dm_scores[$i]==0) {
		    $$lex_dm{$f_phrase}{$e_phrase}[$i]=$unknown_lex_dm_prob;
		} else {
		    $$lex_dm{$f_phrase}{$e_phrase}[$i]=log($dm_scores[$i]);
		}
	    }
	} else {
	    for(my $i=0; $i<@dm_scores; $i++) {
		if($dm_scores[$i]==0) {
		    $$lex_dm{$f_phrase}{$e_phrase}[$i]=$unknown_lex_dm_prob*$lex_dm_feature_weights[$i];
		} else {
		    $$lex_dm{$f_phrase}{$e_phrase}[$i]=log($dm_scores[$i])*$lex_dm_feature_weights[$i];
		}
	    }
	}

    }
    my $time_end=time;
    printf STDERR " done. (%.2f sec)\n", $time_end-$time_begin;

    if(0) {
	print STDERR "Adding entries from phrase table to dm model ... ";
	foreach my $f (keys %$f2e_lex) {
	    foreach my $e (keys %{ $$f2e_lex{$f} }) {
		if(!exists($$lex_dm{$f}) || !exists($$lex_dm{$f}{$e}) || !defined($$lex_dm{$f}{$e}[0])) {
		for(my $i=0; $i<@{ $$lex_dm{'UNK'}{'UNK'} }; $i++) {
		    $$lex_dm{$f}{$e}[$i]=$$lex_dm{'UNK'}{'UNK'}[$i];
		}
		}
	    }
	}
	print STDERR "done.\n";
    }
}

return 1;

