
sub load_phrase_table {

    my($line,@weights);
    my($f_phrase,$e_phrase,$e_pos,$rest);

    my($phrase_file,$f2e_lex_full,$f2e_lex_no_lm,$parameters,$features,$lm_srilm_ids,$nbest_sents,$src_tokens,$rescore_mode,$f2e_lex_features,$nbest_f2e)=@_;
    
    my $poslm_order=$$parameters{'poslm_order'};
    my $lm_order=$$parameters{'lm_order'};
    my $no_lines_processed=0;

    my @phrase_table_feature_weights=@{ $$features{'phrase_table'} };
    my $word_penalty=$$features{'word_penalty'};
    my $cap=$$parameters{'lm_cap'};
    my $phrase_table_threshold=$$parameters{'phrase_table_threshold'};

    my $unknown_word_trans_prob=$$parameters{'unknown_word_trans_prob'};

    my @lm_weights;
    $lm_weights[0]=$$features{'lm_weight'};
    my %nbest_f2e;

    my @empty_array=();

    my %lm_cache;

    my $trg_sorted=$phrase_file=~/\.trg_sorted(\.gz)?$/o ? 1 : 0;

    print STDERR "trg_sorted=$trg_sorted\n";

    if($rescore_mode) {
	for(my $sent_id=0; $sent_id<@$src_tokens; $sent_id++) {
	    my @f_tokens=@{ $src_tokens->[$sent_id] };

#	    print STDERR "nbest_sents->[$sent_id][0]=$nbest_sents->[$sent_id][0]\n";
	    if(!defined($nbest_sents->[$sent_id][0])) {
		print STDERR "Not defined nbest_sents->[$sent_id][0]=$nbest_sents->[$sent_id][0]\n"
	    }
	    for(my $rank=0; $rank<@{ $nbest_sents->[$sent_id] }; $rank++) {
		last if($nbest_sents->[$sent_id][$rank] eq 'NIL');

		my $coverset_string;
		($coverset_string)=$nbest_sents->[$sent_id][$rank]=~/ coversets=([0-9].*)$/o;
		my @covers=split(/\|\|\|/,$coverset_string);
		for(my $i=0; $i<@covers; $i++) {
		    my($l,$r,$e_phrase)=$covers[$i]=~/^([0-9]+):([0-9]+) (.*)$/o;
		    my $f_phrase=join(" ",@f_tokens[$l..$r]);

		    $$nbest_f2e{$f_phrase}{$e_phrase}=1;
		}		
	    }
	}
    }

    my $prev_e_phrase='Z_#';
    my $prev_e_phrase_length=0;
    my $prev_e_phrase_lm_prob=0;

    my $prev_no_lines=0;
    print STDERR "Reading in phrase table \'$phrase_file\' ...  ";
    my $time_begin=time;
    if($phrase_file=~/\.gz$/o) {
	open PHRASETABLE, "<:gzip", $phrase_file, or die("can't open $phrase_file: $!\n");
    } else {
	open(PHRASETABLE,"<$phrase_file") || die("can't open $phrase_file: $!\n");
    }
    while(defined($line=<PHRASETABLE>)) {

	if($line!~/^[\s\t]*\n/o && $line!~/^SENT\_ID\_[0-9]+/o) {

	    chop($line);
	    $no_lines_processed++;
	    my $no_digits = rindex($prev_no_lines,"");
	    if($no_lines_processed % 100000 == 0) {
		for(my $j=1; $j<=$no_digits; $j++) {
		    print STDERR "\x08";
		};
		print STDERR $no_lines_processed;
		$prev_no_lines = $no_lines_processed;
	    }      
	

#	    last if($no_lines_processed>20000);

	    if($poslm_order>0) {
		($f_phrase,$e_phrase,$e_pos,$score_string)=$line=~/^(.+) \|\|\| (.+) \|\|\| (.+) \|\|\|\s+(.+)\n/o;
		undef @e_pos_tags;
		@e_pos_tags=split(/ /,$e_pos);
	    } else {
#		($f_phrase,$e_phrase,$score_string)=$line=~/^(.+) \|\|\| (.+) \|\|\|\s+(.+)\n/;
		if($trg_sorted) {
		    ($e_phrase,$f_phrase,$score_string)=split(/ \|\|\| /o,$line);		    
		} else {
		    ($f_phrase,$e_phrase,$score_string)=split(/ \|\|\| /o,$line);
		}
	    };

	    if($rescore_mode) {
		if(!exists($$nbest_f2e{$f_phrase}) || !exists($$nbest_f2e{$f_phrase}{$e_phrase})) {
		    next;
		}
	    }

	    my @scores=split(/ /,$score_string);
	    my @e_phrase_words;  #=split(' ',$e_phrase);
	    my $e_phrase_length; #=@e_phrase_words;

	    if($phrase_table_threshold) {
		my $skip_entry=0;
		for(my $k=0; $k<4; $k++) {
		    if($scores[$k]<$phrase_table_threshold) {
			$skip_entry=1;
			last;
		    }
		}
		next if($skip_entry);
	    }

	    #substitute numbers:

#	    my @e_phrase_words_lm=@e_phrase_words;
#	    for(my $c=0; $c<@e_phrase_words_lm; $c++) {
##		$e_phrase_words_lm[$c]=~s/[0-9]/6/g;
#		$english_vocab{$e_phrase_words_lm[$c]}=1;
#	    }

#	    my $e_phrase_lm_prob=0;
#	    for(my $lm_id=0; $lm_id<@$lm_srilm_ids; $lm_id++) {
#		$e_phrase_lm_prob+=lm_prob_general($$lm_srilm_ids[$lm_id],\@e_phrase_words_lm,\@empty_array,$cap)*
#		    $lm_weights[$lm_id];
#	    }

	    if($trg_sorted && $e_phrase eq $prev_e_phrase) {
		$e_phrase_length=$prev_e_phrase_length;
		$e_phrase_lm_prob=$prev_e_phrase_lm_prob;
	    } else {
		@e_phrase_words=split(' ',$e_phrase);
		$e_phrase_length=@e_phrase_words;
		$e_phrase_lm_prob=&lm_cost(\@e_phrase_words,\@empty_array,$parameters,$features,$lm_srilm_ids,0);
	    }

	    if($trg_sorted) {
		$prev_e_phrase=$e_phrase;
		$prev_e_phrase_length=$e_phrase_length;
		$prev_e_phrase_lm_prob=$e_phrase_lm_prob;
	    }

	    # the rescore_mode is set to 0 here
#	    my $e_phrase_lm_prob;
#	    if(defined(my $lm_cached=$lm_cache{$e_phrase})) {
#		$e_phrase_lm_prob=$lm_cached;
#	    } else {
#		$e_phrase_lm_prob=&lm_cost(\@e_phrase_words,\@empty_array,$parameters,$features,$lm_srilm_ids,0);
#		$lm_cache{$e_phrase}=$e_phrase_lm_prob;
#	    }

	    my $e_pos_lm_prob=0;
	    if($poslm_order>0) {
		$e_pos_lm_prob=lm_5gram_prob($poslm_id,\@e_pos_tags,$cap)*$poslm_weight;
	    }

	    if($rescore_mode) {
		for(my $i=0; $i<@scores; $i++) {
		    if($scores[$i]==0) {
			$$f2e_lex_features{$f_phrase}{$e_phrase}[$i]=$unknown_word_trans_prob;
		    } else {
			$$f2e_lex_features{$f_phrase}{$e_phrase}[$i]=log($scores[$i]);
		    }
		}
	    };

	    my $trans_prob=0;
	    for(my $i=0; $i<@scores; $i++) {
		if($scores[$i]==0) {
		    $trans_prob+=$unknown_word_trans_prob*$phrase_table_feature_weights[$i];
		} else {		
		    $trans_prob+=log($scores[$i])*$phrase_table_feature_weights[$i];
		}
	    }
	    
	    $trans_prob+=$word_penalty*(-1*$e_phrase_length);
	    $$f2e_lex_no_lm{$f_phrase}{$e_phrase}=$trans_prob;
	    
	    $trans_prob+=$e_phrase_lm_prob;
	    $trans_prob+=$e_pos_lm_prob;
	    
	    $$f2e_lex_full{$f_phrase}{$e_phrase}=$trans_prob;
	    
	    
#	    print STDERR "f2e_lex_full{$f_phrase}{$e_phrase}=$$f2e_lex_full{$f_phrase}{$e_phrase}\n";
	    if($poslm_order>0) {
		$f2e_pos_tmp{$f_phrase}{$e_phrase}=$e_pos;
	    }

	}
    }
    close(FILE);

#    $english_vocab{"<s>"}=1;
#    $english_vocab{"</s>"}=1;

    my $time_end=time;
    printf STDERR " done. (%.2f sec)\n", $time_end-$time_begin;
}

sub restrict_phrase_table {
    my($c);

    my($f2e_lex,$f2e_lex_top_lm,$f2e_lex_full,$f2e_lex_no_lm,$parameters,$lm_srilm_ids,$features,$rescore_mode)=@_;

    my $poslm_order=$$parameters{'poslm_order'};
    my $lm_order=$$parameters{'lm_order'};
    my $translation_limit=$$parameters{'phrase_table_limit'};
    my $translation_beam=log($$parameters{'phrase_table_beam_width'});

    print STDERR "Sorting and restricting phrase table ... ";

    if($poslm_order>0) {
	foreach my $f_phrase (keys %$f2e_lex_full) {
	    $c=1;
	    foreach my $e_phrase (sort {-1*($$f2e_lex_full{$f_phrase}{$a}<=>$$f2e_lex_full{$f_phrase}{$b})} 
				  (keys %{ $$f2e_lex_full{$f_phrase} })) {
		if($c==1) {
		    $$f2e_lex_top_lm{$f_phrase}=$$f2e_lex_full{$f_phrase}{$e_phrase};
		}
		if(!($rescore_mode) && $c>$translation_limit) {
		    last;
		} else {
		    $$f2e_lex{$f_phrase}{$e_phrase}=$$f2e_lex_no_lm{$f_phrase}{$e_phrase};
		    $$f2e_pos{$f_phrase}{$e_phrase}=$$f2e_pos_tmp{$f_phrase}{$e_phrase};
		    $c++;
		}
	    }
	}
    } else {
	foreach my $f_phrase (keys %$f2e_lex_full) {
	    $c=1;
	    foreach my $e_phrase (sort {-1*($$f2e_lex_full{$f_phrase}{$a}<=>$$f2e_lex_full{$f_phrase}{$b})} 
			       (keys %{ $$f2e_lex_full{$f_phrase} })) {
		if($c==1) {
		    $$f2e_lex_top_lm{$f_phrase}=$$f2e_lex_full{$f_phrase}{$e_phrase};
		}
#		if(!($rescore_mode) && ($c>$translation_limit||$$f2e_lex_top_lm{$f_phrase}+$translation_beam>$$f2e_lex_full{$f_phrase}{$e_phrase})) {
		if(!($rescore_mode) && ($c>$translation_limit)) {
		    last;
		} else {
		    $$f2e_lex{$f_phrase}{$e_phrase}=$$f2e_lex_no_lm{$f_phrase}{$e_phrase};	    
		    #add partial LM score here:
		    if(!$rescore_mode) {
			my @e_phrase=split(/ /,$e_phrase);
			if(@e_phrase>=$lm_order) {
			    my @e_phrase_left=@e_phrase[0..($lm_order-2)];
			    my @e_phrase_right=@e_phrase[($lm_order-1)..(@e_phrase-1)];
			    $$f2e_lex{$f_phrase}{$e_phrase}+=&lm_cost(\@e_phrase_left,\@e_phrase_right,$parameters,$features,$lm_srilm_ids,$rescore_mode);
			}
		    }
		    $c++;
		}
	    }
	}
    }

    print STDERR "done.\n";
}


sub add_oov_words_to_phrase_table {
    my($f2e_lex,$f2e_lex_top_lm,$f2e_lex_no_lm,$f_tokens,$features,$default_values,$parameters,$lm_srilm_ids,$rescore_mode,$f2e_lex_features,$nbest_f2e)=@_;

    my @phrase_table_feature_weights=@{ $$features{'phrase_table'} };
    my $word_penalty=$$features{'word_penalty'};
    my $cap=$$parameters{'lm_cap'};
    my @lm_weights;
    $lm_weights[0]=$$features{'lm_weight'};
    my $unknown_word_trans_prob=$$parameters{'unknown_word_trans_prob'};

    my @empty_array=();

    for(my $i=0; $i<@$f_tokens; $i++) {
	if((!$rescore_mode && !exists($$f2e_lex{$f_tokens->[$i]}))
	   || ($rescore_mode && !exists($$f2e_lex{$f_tokens->[$i]}) && exists($$nbest_f2e{$f_tokens->[$i]}))
	    ) {
	    my $prob=0;
	    for(my $j=0; $j<@{ $$default_values{'phrase_table'} }; $j++) {
		$prob+=$$default_values{'phrase_table'}[$j]*$phrase_table_feature_weights[$j];
		if($rescore_mode) {
		    $$f2e_lex_features{$f_tokens->[$i]}{$f_tokens->[$i]}[$j]=$$default_values{'phrase_table'}[$j];
#		    print STDERR "DEFAULT: f2e_lex_features{$f_tokens->[$i]}{$f_tokens->[$i]}[$j]=$$f2e_lex_features{$f_tokens->[$i]}{$f_tokens->[$i]}[$j]\n";
		}
	    }

	    #word_penalty:
	    # non-latin words are deleted later on, so no word penalty for them.
	    if(1 || $f_tokens->[$i]=~/^[0-9A-Za-z\~\`\@\#\$\%\^\&\*\(\)\-\_\=\+\[\{\]\}\\\|\;\:\'\"\,\<\.\>\/\?]+$/) {		
		$prob+=-1*$word_penalty;
	    } else {
#		$$non_latin_oov{$f_tokens->[$i]}=1;
	    }

	    my @e_phrase_words;
	    push(@e_phrase_words,$f_tokens->[$i]);

#	    for(my $c=0; $c<@e_phrase_words_lm; $c++) {
##		$e_phrase_words_lm[$c]=~s/[0-9]/6/g;
#		$english_vocab{$e_phrase_words_lm[$c]}=1;
#	    }
	    
	    # set rescore mode to 0 (last argument)
	   
	    $$f2e_lex{$f_tokens->[$i]}{$f_tokens->[$i]}=$prob;
	    $$f2e_lex_no_lm{$f_tokens->[$i]}{$f_tokens->[$i]}=$prob;

	    $prob+=&lm_cost(\@e_phrase_words,\@empty_array,$parameters,$features,$lm_srilm_ids,0);
	    
#	    $$f2e_lex{$f_tokens->[$i]}{$f_tokens->[$i]}=$prob;
	    $$f2e_lex_top_lm{$f_tokens->[$i]}=$prob;
	}
    }   
}

return 1;
