#!/usr/bin/perl -w

use strict;
use Getopt::Long "GetOptions";
use warnings;

$SIG{CHLD} = 'IGNORE';

BEGIN { unshift @INC, qw( /home/christof/perl/modules/lib64/perl5/site_perl/5.8.8/x86_64-linux-thread-multi ); }
use PerlIO::gzip;
use Devel::Size qw(size total_size);



my $oister_home='/home/christof/scripts/oister';
my $main_dir="$oister_home/src/phrase-based-dm-fork-global";

my $config_file;
my $src_file;
my $trg_file;
my $rescore_mode;
my $rescore_lattice;
my $mert_mode;
my $nbest_file;
my $nbest_size;
my $poslm_order;
our $print_lattice;
my $experiment_dir;
#my $first_sentence;
#my $last_sentence;
#my $batch_id;
my $filter_phrase_table;
my $generate_rescore_lattice;
my $no_parallel;
my $min_batch_size;
my $local_dir;

my %parameters;
my %data;
my %features;
my %feature_order;
my %default_values;

our %f2e_lex;
my %f2e_lex_full;
our %f2e_lex_no_lm;
our %f2e_lex_top_lm;
our %f2e_lex_features;
my %nbest_f2e;
my %max_lm_prob;

my @lm_srilm_ids;
my @src_tokens;
my @nbest_sents;
my @translations;

GetOptions(
    "conf=s" => \$config_file,
    "src=s" => \$src_file,
    "trg=s" => \$trg_file,
    "rescore" => \$rescore_mode,
    "rescore-lattice" => \$rescore_lattice,
    "generate-nolm-lattice" => \$generate_rescore_lattice,
    "nbest-file=s" => \$nbest_file,
    "nbest-size=i" => \$nbest_size,
    "poslm-order=i" => \$poslm_order,
    "print-lattice=i" => \$print_lattice,
    "mert" => \$mert_mode,
    "experiment-dir=s" => \$experiment_dir,
    "no-parallel=i" => \$no_parallel,
    "min-batch-size=i" => \$min_batch_size,
    "use-local-dir=s" => \$local_dir,
#    "first-sentence=i" => \$first_sentence,
#    "last-sentence=i" => \$last_sentence,
#    "batch-id=s" => \$batch_id,
    "filter-phrase-table" => \$filter_phrase_table
    );


print STDERR "conf=$config_file\n";

print STDERR "Unprocessed by Getopt::Long\n" if $ARGV[0];
foreach (@ARGV) {
  print STDERR "$_\n";
}



#default values;
$poslm_order=0 if(!defined($poslm_order));
#$first_sentence=0 if(!defined($first_sentence));
#$last_sentence=10**10 if(!defined($last_sentence));

$mert_mode=defined($mert_mode) ? 1 : 0;
$rescore_mode=defined($rescore_mode) ? 1 : 0;
$generate_rescore_lattice=defined($generate_rescore_lattice) ? 1 : 0;
$rescore_lattice=defined($rescore_lattice) ? 1 : 0;
$filter_phrase_table=defined($filter_phrase_table) ? 1 : 0;
$nbest_size=100 if(!defined($nbest_size));
$min_batch_size||=1;
$local_dir||='';
$local_dir=~s/\/*$//;

$print_lattice=1 if(!defined($print_lattice));
$experiment_dir='.' if(!defined($experiment_dir));
$nbest_file='nil' if(!defined($nbest_file));
$trg_file="$src_file.trans" if(!defined($trg_file));

my $translations_dir="$experiment_dir/translations";
my $lattice_dir="$experiment_dir/lattices";
my $rescored_dir="$experiment_dir/rescored";
my $all_lattices_file="$trg_file.lattices";
my $all_lattices_nolm_file="$trg_file.lattices_nolm";
my $all_rescored_file="$nbest_file.rescored";
#my $batch_mode=defined($batch_id) ? 1 : 0;

my $mode_printout=$rescore_mode?'rescoring':'decoding';

#if($batch_mode) {
#    my ($batch_id_short)=$batch_id=~/finished_batches\/(.+)$/;
#    print STDERR "oister-decoder.pl ($batch_id_short) pid=$$\n";
#} else {
    print STDERR "oister-decoder.pl pid=$$\n";
#}
#
#unlink("$batch_id") if($batch_mode);



print STDERR "rescore_mode=$rescore_mode\n";

if(-e "$lattice_dir" && !$rescore_lattice) {
    print STDERR "rm -rf $lattice_dir\n";
    system("rm -rf $lattice_dir");
} 
system("mkdir $lattice_dir");

if(-e "$rescored_dir") {
    print STDERR "rm -rf $rescored_dir\n";
    system("rm -rf $rescored_dir");
} 
system("mkdir $rescored_dir") if($rescore_mode);

unshift (@INC, "$main_dir");
require 'lm_srilm.pl';
require 'general.pl';
require 'load_phrase_table.pl';
require 'print_progress.pl';
require 'load_config_file.pl';
require 'load_lm_file.pl';
require 'translate_sentence.pl';
require 'find_matching_translations.pl';
require 'future_costs.pl';
require 'initialize_stack.pl';
require 'expand_node.pl';
require 'prune_stack.pl';
require 'return_translation.pl';
require 'print_lattice.pl';
require 'debug.pl';
require 'rescoring.pl';
require 'lm_highest_prob.pl';
require 'batch_filter_phrase_table.pl';
require 'read_lattice_file.pl';
require 'load_lexicalized_distortion_model.pl';

#all the srilm source stuff is under
# /homes/christof/resources/software/swig-srilm-multilm-5gram
use lib '/home/christof/scripts/oister/src/phrase-based-dm-fork-global/srilm';
use srilm;


&load_config_file($config_file,$mert_mode,\%data,\%parameters,\%features,\%feature_order,\%default_values,$rescore_lattice);

if($local_dir ne '') {
    if(!(-e "$local_dir")) {
	print STDERR "mkdir $local_dir\n";
	system("mkdir \"$local_dir\"");
    }
    print STDERR "Using local dir: $local_dir\n";
    foreach my $data_type (keys %data) {
	my $data_file=$data{$data_type};

	my($data_file_name)=$data_file=~/([^\/]+)$/;
	if(-e "$local_dir/$data_file_name") {
	    my $data_orig_size= -s "$data_file";
	    my $data_local_size= -s "$local_dir/$data_file_name";
	    print STDERR "size of $data_file=$data_orig_size\n";
	    print STDERR "size of $local_dir/$data_file_name=$data_local_size\n";
	    if($data_orig_size != $data_local_size) {
		print STDERR "cp \"$data_file\" \"$local_dir/$data_file_name\"\n";
		system("cp \"$data_file\" \"$local_dir/$data_file_name\"");
		$data{$data_type}="$local_dir/$data_file_name";
	    } else {
		print STDERR "Using local version: $local_dir/$data_file_name\n";
	    }
	} else {
	    print STDERR "local file \"$local_dir/$data_file_name\" does not exist.\n";
	    print STDERR "cp \"$data_file\" \"$local_dir/$data_file_name\"\n";
	    system("cp \"$data_file\" \"$local_dir/$data_file_name\"");
	    $data{$data_type}="$local_dir/$data_file_name";
	}
    }
}



my $lm_file=$data{'language_model'};

my $max_lm_prob_file="$lm_file.maxprobs";
$data{'oister_home'}=$oister_home;

if($rescore_lattice) {
    $lm_srilm_ids[0]=&load_lm_file($data{'rescore_language_model'},\%parameters,);
} else {
    $lm_srilm_ids[0]=&load_lm_file($data{'language_model'},\%parameters,);
}

#&compute_highest_lm_probs($lm_file,$max_lm_prob_file,\%features,\%max_lm_prob);


my $max_sents=0; 
my $total_src_tokens=0;
print STDERR "Reading in source file \'$src_file\' ... ";
open(SRC,"<$src_file")||die("can't open src_file: $!\n");
while(defined(my $line=<SRC>)) {
    my @tokens;
    $line=~s/[\s\t]*(.*?)[\s\t]*\n/$1/;
    $line='NIL' if($line=~/^[\s\t]*$/);

    $max_sents++;

    @tokens=split(/ +/,$line);
    push(@src_tokens,[ @tokens ]);
    $total_src_tokens+=@tokens; 
}
close(SRC);
print STDERR "done.\n";

my $batch_size=$total_src_tokens/$no_parallel;
$batch_size=~s/^([0-9]+).*$/$1/;

$batch_size=$min_batch_size if($batch_size<$min_batch_size);

my $current_batch_size=0;
my $start_sentence=0;
my $sentence_id=-1;
my @batches;

open(SRC,"<$src_file")||die("can't open file $src_file: $!\n");
while(defined(my $line=<SRC>)) {
    chomp($line);    
    $sentence_id++;

    my @tokens=split(/ +/,$line);
    $current_batch_size+=@tokens;

    if($current_batch_size>=$batch_size) {
        push(@batches,"$start_sentence $sentence_id");
        $start_sentence=$sentence_id+1;
        $current_batch_size=0;
    }
}
close(SRC);
my $last_sent_id=$sentence_id;

#not sure about this one:
if($sentence_id>$start_sentence) {
    push(@batches,"$start_sentence $sentence_id");
}


if($rescore_mode) {
    open(NBEST,"<$nbest_file")||die("can't open nbest_file: $!\n");
    my $sent_id;
    while(defined(my $line=<NBEST>)) {
	if($line=~/^<SENT ID=([0-9]+)>/) {
	    $sent_id=$1;
#	    $sent_id--;	    
	} else {
	    while($line!~/^<\/SENT>/) {
		chomp($line);
		push(@{ $nbest_sents[$sent_id] },$line);
		last if(!defined($line=<NBEST>));
	    }
	}
    }
    close(NBEST);
} else {
    for(my $sent_id=0; $sent_id<@src_tokens; $sent_id++) {
	push(@{ $nbest_sents[$sent_id] }, 'NIL');
    }
}


#if($filter_phrase_table && $data{'phrase_table'}!~/\.trg_sorted(\.gz)$/) {
#    $data{'phrase_table'}=&batch_filter_phrase_table(\%data,\%parameters,$experiment_dir,\@src_tokens,$first_sentence,$last_sentence);
#}

&load_phrase_table($data{'phrase_table'},\%f2e_lex_full,\%f2e_lex_no_lm,\%parameters,\%features,\@lm_srilm_ids,\@nbest_sents,\@src_tokens,$rescore_mode,\%f2e_lex_features,\%nbest_f2e);


&restrict_phrase_table(\%f2e_lex,\%f2e_lex_top_lm,\%f2e_lex_full,\%f2e_lex_no_lm,\%parameters,\@lm_srilm_ids,\%features,$rescore_mode);

undef %f2e_lex_full;
undef %f2e_lex_no_lm if(!$generate_rescore_lattice);

print STDERR "Adding OOV words to phrase-table ... ";
for(my $i=0; $i<@src_tokens; $i++) {
    &add_oov_words_to_phrase_table(\%f2e_lex,\%f2e_lex_top_lm,\%f2e_lex_no_lm,\@{ $src_tokens[$i] },\%features,\%default_values,\%parameters,\@lm_srilm_ids,$rescore_mode,\%f2e_lex_features,\%nbest_f2e);
}
print STDERR "done.\n";

our %lex_dm;
&load_lexicalized_distortion_model($data{'lex_dm_model'},\%lex_dm,\%f2e_lex,\%parameters,\%features,$rescore_mode);

#my @total_start_times = times;

#my %children_forks;
my @children_forks;
for(my $batch=0; $batch<@batches; $batch++) {
    my $pid=fork();
    print STDERR "Result of forking batch $batch: $!\n";

    if($pid) {
	push(@children_forks,$pid);
#	$children_forks{$pid}++;
    } elsif($pid==0) {
	my @total_start_times = times;

	my($first_sentence,$last_sentence)=split(/ /,$batches[$batch]);
 	print STDERR "Forking off batch $batch (sentences $first_sentence-$last_sentence) pid=$$\n";	
	open(T,">$experiment_dir/translations.$batch");
	if($rescore_mode) {
	    open(R,">$all_rescored_file.batch.$batch");
	}

	for(my $sent_id=$first_sentence; $sent_id<@src_tokens && $sent_id<=$last_sentence; $sent_id++) {
 
	    my @start_times = times;
	    my @rescored_sents=();
	    my %lm_cache;
	    undef %lm_cache;
	    my %lm_between_cache;
	    undef %lm_between_cache;

	    my @sentence_tokens=@{ $src_tokens[$sent_id] };

	    my $rescore_lattice_sentence_file=$rescore_lattice ? "$experiment_dir/lattices/$sent_id.lattice.nolm" : '';

#   print STDERR "sent_id=$sent_id\n";
	    for(my $nbest_rank=0; $nbest_rank<@{ $nbest_sents[$sent_id] }; $nbest_rank++) {
#       print STDERR "nbest_rank=$nbest_rank\n";
#       print STDERR "nbest_sents[$sent_id][$nbest_rank]=$nbest_sents[$sent_id][$nbest_rank]=\n";
		my $translation=&translate_sentence(\@sentence_tokens,$rescore_mode,$generate_rescore_lattice,$rescore_lattice,\$rescore_lattice_sentence_file,\$nbest_sents[$sent_id][$nbest_rank],$nbest_rank,\%parameters,\%features,\%feature_order,\@lm_srilm_ids,$sent_id,$lattice_dir,\@rescored_sents,\%lm_cache,\%lm_between_cache);
		
		if(!$rescore_mode) {
#		    $translations[$sent_id]=$translation;
		    print T $translation, "\n";

		}
	    }
	    undef %lm_cache;
	    undef %lm_between_cache;
	    
	    if($rescore_mode) {
		for(my $i=0; $i<@rescored_sents; $i++) {
		    print R "$sent_id ||| $rescored_sents[$i]\n";
		}
#		&print_rescored_sentences($sent_id,\@rescored_sents,$rescored_dir);
	    }
	    my @end_times = times;

	    printf STDERR "sent=$sent_id CPU sec=%.2f\n", $end_times[0]-$start_times[0];
	}
	close(T);
	my @total_end_times = times;

	close(R) if($rescore_mode);

	printf STDERR "total $mode_printout time for batch $batch: CPU sec=%.2f\n", $total_end_times[0]-$total_start_times[0];
	my $date=`date`;
	print STDERR "batch $batch ($mode_printout) finished at $date\n";

	open(TIME,">$experiment_dir/$mode_printout.time.$batch");
	printf TIME "total $mode_printout time: CPU sec=%.2f\n", $total_end_times[0]-$total_start_times[0];
	close(TIME);
	
	exit(0);    
    } else {
	die("Couldn't fork: $!\n");
    }
}
#my @total_end_times = times;

foreach my $child (@children_forks) {
    waitpid($child,0);
}


my $date=`date`;
print STDERR "All children have finished. $date\n";

#while (%children_forks) {
#   my $child_pid = waitpid(-1, 0);
#   delete $children_forks{$child_pid};
#}


#if(!$rescore_mode) {
#    printf STDERR "total decoding time: CPU sec=%.2f\n", $total_end_times[0]-$total_start_times[0];
#} elsif($rescore_mode) {
#    printf STDERR "total rescoring time: CPU sec=%.2f\n", $total_end_times[0]-$total_start_times[0];
#}

if(!$rescore_mode) {
    unlink("$trg_file") if(-e "$trg_file");
    system("touch $trg_file");
    for(my $batch=0; $batch<@batches; $batch++) {
	print STDERR "cat $experiment_dir/translations.$batch >> $trg_file\n";
	system("cat $experiment_dir/translations.$batch >> $trg_file");
	unlink("$experiment_dir/translations.$batch") if(-e "$experiment_dir/translations.$batch");
    }


    print STDERR "rm -f $all_lattices_file.batch.*\n";
    system("rm -f $all_lattices_file.batch.*");
    for(my $batch=0; $batch<@batches; $batch++) {
	my($first_sentence,$last_sentence)=split(/ /,$batches[$batch]);
#	unlink "$all_lattices_file.batch.$batch" if(-e "$all_lattices_file.batch.$batch");

	open(ALL,">$all_lattices_file.batch.$batch");
	for(my $sent_id=$first_sentence; $sent_id<@src_tokens && $sent_id<=$last_sentence; $sent_id++) {
	    open(INDLATTICE,"<$lattice_dir/$sent_id.lattice");
	    while(defined(my $line=<INDLATTICE>)) {
		print ALL $line;
	    }
	    close(INDLATTICE);
	    unlink("$lattice_dir/$sent_id.lattice");
	}
	close(ALL);
    }
}


if(0 && $generate_rescore_lattice) {
    open(ALL,">$all_lattices_nolm_file");
    for(my $batch=0; $batch<@batches; $batch++) {
	my($first_sentence,$last_sentence)=split(/ /,$batches[$batch]);
	for(my $sent_id=$first_sentence; $sent_id<@src_tokens && $sent_id<=$last_sentence; $sent_id++) {
	    open(INDLATTICE,"<$lattice_dir/$sent_id.lattice.nolm");
	    while(defined(my $line=<INDLATTICE>)) {
		print ALL $line;
	    }
	    close(INDLATTICE);
	    unlink("$lattice_dir/$sent_id.lattice.nolm");
	}
    }
    close(ALL);
}


if(0 && $rescore_mode) {
    unlink "$all_rescored_file" if(-e "$all_rescored_file");
    system("touch $all_rescored_file");
    for(my $sent_id=0; $sent_id<@src_tokens; $sent_id++) {
	system("cat $rescored_dir/$sent_id.rescored >> $all_rescored_file");
    }

    for(my $sent_id=0; $sent_id<@src_tokens; $sent_id++) {
	unlink("$rescored_dir/$sent_id.rescored");
    }
    system("rmdir $rescored_dir");    
}

my $avg_batch_time=0;
my $total_batch_time=0;
my $min_batch_time=10**10;
my $max_batch_time=0;
for(my $batch=0; $batch<@batches; $batch++) {
    open(E,"<$mode_printout.time.$batch");
    my $time_line=<E>;
    my($batch_time)=$time_line=~/^total $mode_printout time: CPU sec=([0-9\.e\-]+)\n/;
    $avg_batch_time+=$batch_time;
    $min_batch_time=$batch_time if($batch_time<=$min_batch_time);
    $max_batch_time=$batch_time if($batch_time>=$max_batch_time);
    close(E);
    unlink("$mode_printout.time.$batch");
}
$total_batch_time=$avg_batch_time;
$avg_batch_time/=@batches;

printf STDERR "total $mode_printout time: CPU sec=%.2f\n", $total_batch_time;
printf STDERR "average $mode_printout time: CPU sec=%.2f\n", $avg_batch_time;
printf STDERR "min $mode_printout time: CPU sec=%.2f\n", $min_batch_time;
printf STDERR "max $mode_printout time: CPU sec=%.2f\n", $max_batch_time;

